"""File handler implementation."""
import io
import os
import uuid
import json
import shutil
import zipfile
from pathlib import Path
from typing import Optional

import requests

from lpmn_client_biz import Connection, IOType

UPLOADED = 'Uploaded'
LPMN_CLIENT_DIR = 'lpmn_client_input'
FILENAME = 'filename'

FileID = str

requests.packages.urllib3.disable_warnings(
    requests.packages.urllib3.exceptions.InsecureRequestWarning)


def jsonl_to_directory(input_path, column_name):
    """Converts a jsonl file to a directory with files.

    :param input_path: path to the input jsonl file
    :type input_path: str
    :param column_name: name of the column to extract from the jsonl file
    :type column_name: str
    :return: path to the output directory
    :rtype: str
    """
    output_path = str(uuid.uuid4())
    with open(input_path, 'r') as fin:
        json_files = [json.loads(obj) for obj in fin.readlines()]
    os.makedirs(output_path)
    for idx, json_file in enumerate(json_files):
        if column_name in json_file:
            # json object case
            text = str(json_file[column_name])
            with open(f'{output_path}/file_{idx}',
                      'w', encoding='utf8') as fout:
                fout.write(text)
    return output_path


def archive_to_jsonl(
        archive_path,
        output_file,
        json_file,
        columns_to_take=[]
):
    """Appended archive desired columns to json file.

    :param json_file: path to the jsonl file
    :type json_file: str
    :param columns_to_take: list of columns to take from the archive files
    :type columns_to_take: list
    :param archive_path: path to the zip archive
    :type archive_path: str
    :param output_file: path to the output jsonl file
    :type output_file: str
    """
    output_path = archive_path.replace('.zip', '')
    shutil.unpack_archive(archive_path, output_path)
    with open(json_file, 'r') as fin:
        json_lines = [json.loads(obj) for obj in fin.readlines()]
    files = sorted(os.listdir(output_path), key=lambda x: int(x.split('_')[1]))
    for j, filename in enumerate(files):
        with open(f'{output_path}/{filename}', 'r') as file:
            lines = [json.loads(line) for line in file.readlines()]
            if len(lines) > 0:
                joined_json = {}
                for idx, line in enumerate(lines):
                    if idx == 0:
                        joined_json[FILENAME] = line[FILENAME].replace('_0', '')
                        for col in columns_to_take:
                            if col in line:
                                joined_json[col] = line[col]
                    else:
                        for col in columns_to_take:
                            if col in line:
                                if isinstance(line[col], list):
                                    joined_json[col].extend(line[col])
                                elif isinstance(line[col], str):
                                    joined_json[col] += line[col]
                                joined_json[col] = line[col]
            else:
                joined_json = lines[0]
            json_lines[j].update(
                {
                    key: value
                    for key, value in joined_json.items()
                    if key in columns_to_take
                }
            )
    with open(output_file, 'w') as fd:
        for entry in json_lines:
            json.dump(entry, fd)
            fd.write('\n')
    os.remove(archive_path)


def file_generator(dirname: str) -> str:
    """Yields all file paths from directory.

    :param dirname: path to directory
    :type: dirname: str
    :rtype: str
    """
    for root, dirs, files in os.walk(dirname):
        for name in files:
            root_path = os.path.join(root, name)
            yield root_path, os.path.relpath(root_path, dirname)


def extract_files(file_path: str):
    """Extract files from directory or return file.

    :param file_path: path to file or directory
    :type file_path: str
    :return: file or zip buffer
    :rtype: IOType
    """
    if os.path.isdir(file_path):
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(
                zip_buffer,
                mode="a",
                compression=zipfile.ZIP_DEFLATED,
        ) as zip_file:
            for filename_value, arcname in file_generator(file_path):
                zip_file.write(filename_value, arcname)
        zip_buffer.seek(0)
        files = zip_buffer
    else:
        files = open(file_path, 'rb')
    return files


def upload(connection: Connection, file_path: str) -> FileID:
    """Uploading file to server.

    :param connection: instance of connection class
    :type connection: Connection
    :param file_path: path to local file that should be
        sent to server
    :type file_path: str
    :return: id of file located on server
    :rtype: FileID
    """
    filename = f'{LPMN_CLIENT_DIR}/{uuid.uuid4()}'
    url = f"{connection.base_url}{connection.API_ENDPOINTS['files']}{filename}"
    files = extract_files(file_path)

    r = requests.post(
        url,
        files={
            "f": (filename, files,
                  "application/octet-stream")
        },
        headers=connection.headers(),
        verify=False
    )
    if r.status_code >= 400 and filename in r.json():
        raise requests.HTTPError(f'{r.status_code} '
                                 f'Client Error: {r.json()[filename]}')
    else:
        r.raise_for_status()
    data = r.json()
    if data[filename] == UPLOADED:
        return filename


def delete(connection: Connection, file_id) -> str:
    """Deleting file from server.

    :param connection: instance of Connection class
    :type connection: Connection
    :param file_id: id of file located on server
    :type file_id: FileID
    :return: deleted file id
    :rtype: str
    """
    r = requests.delete(
        f"{connection.base_url}{connection.API_ENDPOINTS['files']}{file_id}",
        headers=connection.headers(),
        verify=False
    )
    r.raise_for_status()
    if r.json() == file_id:
        return file_id


def download(connection: Connection,
             file_id: FileID,
             output_type: IOType,
             output_path: Optional[str] = None,
             filename: Optional[str] = None
             ) -> str:
    """Downloading file from server.

    :param connection: instance of Connection class
    :type connection: Connection
    :param output_type: preferred type of output
    :type output_type: IOType
    :param file_id: id of file located on server
    :type file_id: FileID
    :param output_path: path to local dir
        where file should be downloaded
        if not provided value is current working directory
    :type output_path: str
    :param filename: target filename of downloaded file
        if not provided value is file_id
    :return: downloaded file path
    :rtype: str
    """
    response_list = requests.get(
        url=f"{connection.base_url}/files/{file_id}?mode=list",
        headers=connection.headers(),
        verify=False
    )
    response_list.raise_for_status()
    files = response_list.json()

    r = requests.get(
        url=f"{connection.base_url}/files/{file_id}?mode=download",
        headers=connection.headers(),
        verify=False
    )
    r.raise_for_status()

    if not filename:
        filename = os.path.basename(file_id)
    if not output_path:
        output_path = os.getcwd()
    os.makedirs(output_path, exist_ok=True)

    if len(files) > 1 or files[0].endswith("/"):
        filename = f"{Path(filename).stem}.zip"

    if output_type == IOType.TEXT:
        return r.content
    else:
        output = f"{output_path}/{filename}"
        with open(output, "wb") as file:
            file.write(r.content)

        return output
